/*
  $Id: LogStructMem.cpp,v 1.5 2016/08/31 08:45:13 ewan Exp $
  Log structured memory Arduino library with wear levelling.
  LogStructMem.cpp (C) 2016 Ewan Parker.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

  The author may be reached at http//www.ewan.cc/ for queries.
*/

#include <Arduino.h>
#include "LogStructMem.h"

// Flag the state of the data in each header.  The state will change from CLEAN
// to INFLUX to VALID by clearing each MSB.
#define LOGMEM_ENT_FLAG_CLEAN  0xFF
#ifdef LOGMEM_CFG_2PHASE
#define LOGMEM_ENT_FLAG_INFLUX 0x7F
#endif
#define LOGMEM_ENT_FLAG_VALID  0x3F

// Create a new log structured memory instance.  Capture the type of the memory
// from which we can find the total size and other properties.  We also need to
// know the size of each block of user data.
LogStructMem::LogStructMem(MemImpl *memory, uint8_t entrySize)
{
  _mem = memory;
  _len = entrySize;
  _alen = _align_page_size(sizeof(_head) + _len);
}

// Ensure we write at a page boundary by rounding up to a multiple of bits.
// Return the parameter, page aligned.
uint8_t LogStructMem::_align_page_size(uint8_t s)
{
  uint8_t m = _mem->getBoundaryMask();
  if ((s & m) != 0) return (s | m) + 1;
  else return s;
}

// Scan through memory and return the earliest data.  Return 'false' if nothing
// was found.
boolean LogStructMem::readEarliest(void *data)
{
  #ifdef LOGMEM_CFG_DEBUG
  Serial.print("rdEarliest :: sz="); Serial.print(_mem->getMemSize());
  Serial.print(" len="); Serial.print(_len);
  Serial.print(" alen="); Serial.print(_alen);
  Serial.println(); Serial.flush();
  #endif

  // Count backwards from the latest entry.
  if (readLatest(data))
  {
    uint16_t s = _head.seq - 1;
    boolean done = false;
    uint8_t wraps = 0;
    uint32_t final_addr = (_mem->getMemSize()/_alen - 1) * _alen;
    // Currently we do not write data to the end of the memory space.
    if (final_addr + _alen == _mem->getMemSize()) final_addr -= _alen;
    while (!done) {
      if (_addr == 0)
      {
         _addr = final_addr;
         ++wraps;
      }
      else _addr -= _head.size;
      _mem->readBlock(&_head, _addr, sizeof(_head));

      #ifdef LOGMEM_CFG_DEBUG
      Serial.print("CURR<-- ::");
      Serial.print(" prv_seq="); Serial.print(s);
      Serial.print(" addr="); Serial.print(_addr);
      Serial.print(" flag=0x"); Serial.print(_head.flag, HEX);
      Serial.print(" size="); Serial.print(_head.size);
      Serial.print(" real_seq="); Serial.print(_head.seq);
      Serial.print(" wraps="); Serial.print(wraps);
      Serial.println(); Serial.flush();
      #endif

      if (_head.flag != LOGMEM_ENT_FLAG_VALID
      || (_addr == final_addr && wraps >= 2)
      || _head.size != _alen
      || (_head.seq != s && _head.flag == LOGMEM_ENT_FLAG_VALID))
      {
        // Either at the end of written memory (CLEAN) or at a half-written
        // entry (INFLUX) or wrapped around (VALID or second iteration through
        //  memory), or an unexpected size or sequence.
        if (_addr != final_addr) // The earliest valid entry is the next item.
          _addr += _alen;
        else // The earliest entry is at the start of the memory.
          _addr = 0;
        _mem->readBlock(&_head, _addr, sizeof(_head));
        _mem->readBlock(data, _addr + sizeof(_head), _len);
        _addr += _alen; ++_head.seq;
        if (_addr + _alen >= _mem->getMemSize()) _addr = 0; // Wrap around.
        done = true;
      }
      else --s;
    }
    return true;
  }
  else return false;
}

// Return the next block of data.  Return 'false' if there is no more data.
boolean LogStructMem::readNext(void *data)
{
  #ifdef LOGMEM_CFG_DEBUG
  Serial.print("rdNext :: sz="); Serial.print(_mem->getMemSize());
  Serial.print(" addr="); Serial.print(_addr);
  Serial.print(" seq="); Serial.print(_head.seq);
  Serial.println(); Serial.flush();
  #endif

  uint16_t s = _head.seq;
  _mem->readBlock(&_head, _addr, sizeof(_head));

  #ifdef LOGMEM_CFG_DEBUG
  Serial.print("NEXT--> ::");
  Serial.print(" nxt_seq="); Serial.print(s);
  Serial.print(" addr="); Serial.print(_addr);
  Serial.print(" flag=0x"); Serial.print(_head.flag, HEX);
  Serial.print(" size="); Serial.print(_head.size);
  Serial.print(" real seq="); Serial.print(_head.seq);
  Serial.println(); Serial.flush();
  #endif

  if (_head.flag != LOGMEM_ENT_FLAG_VALID
  || _head.size != _alen
  || (_head.seq != s && _head.flag == LOGMEM_ENT_FLAG_VALID))
  {
    // Either at the end of written memory (CLEAN) or at a half-written
    // entry (INFLUX) or wrapped around (VALID), or an unexpected size or
    // sequence.
    memset(&_head, 0, sizeof(_head));
    _head.flag = LOGMEM_ENT_FLAG_CLEAN;
    _head.seq = s;
    memset(data, 0, _len);
    return false;
  }
  else
  {
    _mem->readBlock(data, _addr + sizeof(_head), _len);
    ++_head.seq; _addr += _alen;
    if (_addr + _alen >= _mem->getMemSize()) _addr = 0; // Wrap around.
    return true;
  }
}

// Scan through memory and return the latest data.  Return 'false' if nothing
// was found.
boolean LogStructMem::readLatest(void *data)
{
  #ifdef LOGMEM_CFG_DEBUG
  Serial.print("rdLatest :: sz="); Serial.print(_mem->getMemSize());
  Serial.print(" len="); Serial.print(_len);
  Serial.print(" alen="); Serial.print(_alen);
  Serial.println(); Serial.flush();
  #endif

  _addr = 0;
  uint16_t s;
  boolean done = false;
  while (!done) {
    _mem->readBlock(&_head, _addr, sizeof(_head));
    if (_addr == 0) s = _head.seq;

    #ifdef LOGMEM_CFG_DEBUG
    Serial.print("CURR--> ::");
    Serial.print(" nxt_seq="); Serial.print(s);
    Serial.print(" addr="); Serial.print(_addr);
    Serial.print(" flag=0x"); Serial.print(_head.flag, HEX);
    Serial.print(" size="); Serial.print(_head.size);
    Serial.print(" real_seq="); Serial.print(_head.seq);
    Serial.println(); Serial.flush();
    #endif

    if (_head.flag != LOGMEM_ENT_FLAG_VALID
    || _addr + _alen >= _mem->getMemSize()
    || _head.size != _alen
    || (_head.seq != s && _head.flag == LOGMEM_ENT_FLAG_VALID))
    {
      // Either at the end of written memory (CLEAN) or at a half-written
      // entry (INFLUX) or wrapped around (VALID), or an unexpected size or
      // sequence.
      if (_addr != 0) { // The latest valid entry was the previously read item.
        _mem->readBlock(&_head, _addr - _alen, sizeof(_head));
        _mem->readBlock(data, _addr - _alen + sizeof(_head), _len);
        ++_head.seq;
        if (_addr + _alen >= _mem->getMemSize()) _addr = 0; // Wrap around.
      }
      else { // No entries exist.
        memset(&_head, 0, sizeof(_head));
        _head.flag = LOGMEM_ENT_FLAG_CLEAN;
        memset(data, 0, _len);
      }
      done = true;
    }
    else { _addr += _alen; ++s; }
  }
  return _head.flag == LOGMEM_ENT_FLAG_VALID;
}

// Write a new entry (replacing the previous entry).  You will probably wish
// to seek to the correct place in memory before calling this (e.g. by
// calling readLatest(void*) beforehand).
void LogStructMem::writeNew(void *data)
{
  _head.size = _alen;
  #ifdef LOGMEM_CFG_2PHASE
  _head.flag = LOGMEM_ENT_FLAG_INFLUX;
  #else
  _head.flag = LOGMEM_ENT_FLAG_VALID;
  #endif
  _mem->writeBlock(&_head, _addr, sizeof(_head));
  _mem->writeBlock(data, _addr + sizeof(_head), _len);
  #ifdef LOGMEM_CFG_2PHASE
  _head.flag = LOGMEM_ENT_FLAG_VALID;
  _mem->writeBlock(&_head, _addr, sizeof(_head));
  #endif
  _addr += _head.size;
  if (_addr + _head.size >= _mem->getMemSize()) // Wrap around.
    _addr = 0;
  ++_head.seq;
}
