/*
  $Id: MemImplAT24C32.cpp,v 1.1 2016/08/27 07:55:03 ewan Exp $
  Atmel AT24C32 EEPROM specific memory implementation.
  MemImplAT24C32.cpp (C) 2016 Ewan Parker.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

  The author may be reached at http//www.ewan.cc/ for queries.
*/

#include <Arduino.h>
#include <Wire.h>
#include <MemImplAT24C32.h>

void acknowledgementPolling(void);

// Constructor.
MemImplAT24C32::MemImplAT24C32()
{
  Wire.begin();
}

// Total size of the memory.  4kB (4096x8) for Atmel AT24C32 EEPROM.
uint32_t MemImplAT24C32::getMemSize()
{
  return 4096;
}

// Page boundary mask for alignment of page reads and writes.  32 bytes for
// Atmel AT24C32 EEPROM.
uint8_t MemImplAT24C32::getBoundaryMask()
{
  return 0x1F;
}

// Read a block from memory at address mem to *data with length len.
void MemImplAT24C32::readBlock(void *data, uint32_t mem, uint8_t len)
{
  #ifdef AT24C32_CFG_DEBUG
  Serial.print("rdBlk data="); Serial.print((uint32_t)data);
  Serial.print(" mem="); Serial.print(mem);
  Serial.print(" len="); Serial.print(len);
  Serial.println();
  #endif

  char *addr = (char *)data;
  Wire.beginTransmission(AT24C32_I2C_ADDR);
  Wire.write(mem >> 8); Wire.write(mem & 0xFF);
  Wire.endTransmission();
  Wire.requestFrom((uint8_t)AT24C32_I2C_ADDR, len);
  for (uint8_t i = 0; i < len; i++) *addr++ = Wire.read();
}

// Write a block to memory at address mem from *data with length len.
void MemImplAT24C32::writeBlock(void *data, uint32_t mem, uint8_t len)
{
  #ifdef AT24C32_CFG_DEBUG
  Serial.print("wrBlk data="); Serial.print((uint32_t)data);
  Serial.print(" mem="); Serial.print(mem);
  Serial.print(" len="); Serial.print(len);
  Serial.println();
  #endif

  Wire.beginTransmission(AT24C32_I2C_ADDR);
  Wire.write(mem >> 8); Wire.write(mem & 0xFF);
  Wire.write((uint8_t *)data, len);
  Wire.endTransmission();
  // Acknowledgement polling following the write operation.
  acknowledgementPolling();

  #ifdef AT24C32_CFG_VERIFY
  Wire.beginTransmission(AT24C32_I2C_ADDR);
  Wire.write(mem >> 8); Wire.write(mem & 0xFF);
  Wire.endTransmission();
  boolean okay = true; int i;
  uint8_t *addr = (uint8_t *)data;
  Wire.requestFrom((uint8_t)AT24C32_I2C_ADDR, len);
  for (i = 0; i < len && okay; i++) if (*addr++ != Wire.read()) okay = false;
  #ifdef AT24C32_CFG_DEBUG
  Serial.print("v:"); Serial.print(okay);
  Serial.print(" i="); Serial.println(i);
  #endif /*AT24C32_CFG_DEBUG*/
  if (!okay) writeBlock(data, mem, len); // Rewrite block.
  #endif /*AT24C32_CFG_VERIFY*/
}

// Check any in-flight operation on flash has completed before returning.
void acknowledgementPolling(void)
{
  Wire.beginTransmission(AT24C32_I2C_ADDR);
  while(Wire.endTransmission()) Wire.beginTransmission(AT24C32_I2C_ADDR);
}
