/*
  $Id: MemImplW25Q64FV.cpp,v 1.1 2016/08/28 16:11:44 ewan Exp $
  Winbond W25Q64FV flash specific memory implementation.
  MemImplW25Q64FV.cpp (C) 2016 Ewan Parker.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

  The author may be reached at http//www.ewan.cc/ for queries.
*/

#include <Arduino.h>
#include <SPI.h>
#include <MemImplW25Q64FV.h>

// IC constants.
#define W25_CONF_CAPACITY        0x800000
#define W25_CONF_PAGE_SIZE       0x100
#define W25_CONF_SPI_SPEED       104000000

// Commands without side effects (read only).
#define W25_CMD_ENABLE_RESET     0x66
#define W25_CMD_RESET            0x99
#define W25_CMD_JEDEC_ID         0x9F
#define W25_CMD_READ_STATUS_REG1 0x05
#define W25_CMD_READ_STATUS_REG2 0x35
#define W25_CMD_READ_ADDRESS     0x03

// Commands with side effects (potentially erase/write).
#define W25_CMD_WRITE_ENABLE     0x06
#define W25_CMD_CHIP_ERASE       0xC7
#define W25_CMD_SECTOR_ERASE_4KB 0x20
#define W25_CMD_PAGE_PROGRAM     0x02

SPISettings spiSet = SPISettings(W25_CONF_SPI_SPEED, MSBFIRST, SPI_MODE0);

// Constructor.
MemImplW25Q64FV::MemImplW25Q64FV(uint8_t csPin)
{
  _cs = csPin;

  pinMode(csPin, OUTPUT);
  SPI.begin();

  SPI.beginTransaction(spiSet);
  _chipSelect();
  SPI.transfer(W25_CMD_ENABLE_RESET);
  _chipDeSelect();
  _chipSelect();
  SPI.transfer(W25_CMD_RESET);
  _chipDeSelect();
  SPI.endTransaction();
  //delay(1); // delay for 30 us
  _waitChipReady();
}

// Total size of the memory.  8MB for Winbond W25Q64FV flash.
uint32_t MemImplW25Q64FV::getMemSize()
{
  return W25_CONF_CAPACITY;
}

// Page boundary mask for alignment of page reads and writes.  N/A for Winbond
// W25Q64FV flash.
uint8_t MemImplW25Q64FV::getBoundaryMask()
{
  return 0x00;
}

void MemImplW25Q64FV::_chipSelect(void)
{
  digitalWrite(_cs, LOW);
}

void MemImplW25Q64FV::_chipDeSelect(void)
{
  digitalWrite(_cs, HIGH);
}

boolean MemImplW25Q64FV::_chipBusy(void)
{
  uint8_t s1, s2;
  SPI.beginTransaction(spiSet);
  _chipSelect();
  SPI.transfer(W25_CMD_READ_STATUS_REG1);
  s1 = SPI.transfer(0x00);
  _chipDeSelect();
  _chipSelect();
  SPI.transfer(W25_CMD_READ_STATUS_REG2);
  s2 = SPI.transfer(0x00);
  _chipDeSelect();
  SPI.endTransaction();
  #ifdef W25Q64FV_CFG_DEBUG
  Serial.print("Reg1/2="); Serial.print(s1, HEX);
  Serial.print("/"); Serial.print(s2, HEX);
  Serial.print(" BUSY="); Serial.print(s1 & 1);
  Serial.print(" WEL="); Serial.print(s1 & 2);
  Serial.print(" SUS="); Serial.print(s2 & 128);
  #endif
  return s1 & 1;
}

void MemImplW25Q64FV::_waitChipReady(void)
{
  #ifdef W25Q64FV_CFG_DEBUG
  Serial.print("Waiting ");
  #endif
  while (_chipBusy()) {
    #ifdef W25Q64FV_CFG_DEBUG
    Serial.print('.');
    #endif
  }
  #ifdef W25Q64FV_CFG_DEBUG
  Serial.println(" DONE");
  #endif
}

void MemImplW25Q64FV::_sendSeek(uint32_t mem)
{
  uint8_t addressByte, shift = 16;
  for (int i = 1; i <= 3; i++)
  {
    addressByte = (mem >> shift) & 0xFF;
    SPI.transfer(addressByte);
    shift -= 8;
  }
}

void MemImplW25Q64FV::_erase4kSector(uint32_t mem)
{
  #ifdef W25Q64FV_CFG_DEBUG
  Serial.print("erSec addr="); Serial.print(mem);
  Serial.println();
  #endif
  SPI.beginTransaction(spiSet);
  _chipSelect();
  SPI.transfer(W25_CMD_WRITE_ENABLE);
  _chipDeSelect();
  _chipSelect();
  SPI.transfer(W25_CMD_SECTOR_ERASE_4KB);
  _sendSeek(mem);
  _chipDeSelect();
  SPI.endTransaction();
  _waitChipReady();
}

// Read a block from memory at address mem to *data with length len.
void MemImplW25Q64FV::readBlock(void *data, uint32_t mem, uint8_t len)
{
  #ifdef W25Q64FV_CFG_DEBUG
  Serial.print("rdBlk data="); Serial.print((uint32_t)data);
  Serial.print(" mem="); Serial.print(mem);
  Serial.print(" len="); Serial.print(len);
  Serial.println();
  #endif

  SPI.beginTransaction(spiSet);
  _chipSelect();
  SPI.transfer(W25_CMD_READ_ADDRESS);
  _sendSeek(mem);

  uint8_t *addr = (uint8_t *)data;
  for (int i = 0; i < len; i++) *addr++ = SPI.transfer(0x00);
  _chipDeSelect();
  SPI.endTransaction();
  _waitChipReady();
}

// Write a block to memory at address mem from *data with length len.
void MemImplW25Q64FV::writeBlock(void *data, uint32_t mem, uint8_t len)
{
  #ifdef W25Q64FV_CFG_DEBUG
  Serial.print("wrBlk data="); Serial.print((uint32_t)data);
  Serial.print(" mem="); Serial.print(mem);
  Serial.print(" len="); Serial.print(len);
  Serial.println();
  #endif

  // If we are about to write over a page boundary then write the partial data
  // and seek to the next page to write the remainder.
  uint32_t page1 = mem >> 8, page2 = (mem + len - 1) >> 8;
  if (page1 != page2)
  {
    uint8_t len1 = 0x100 - (mem & 0xFF);
    uint8_t len2 = len - len1;
    #ifdef W25Q64FV_CFG_DEBUG
    Serial.print("Split len1="); Serial.print(len1);
    Serial.print(" len2="); Serial.print(len2);
    Serial.println();
    #endif
    writeBlock(data, mem, len1);
    writeBlock(data + len1, mem + len1, len2);
  }
  else
  {
    //// If we are going to write into the next sector, then erase it now.
    //uint32_t sec1 = mem >> 12, sec2 = (mem + len) >> 12;
    //if ((mem & 0x0FFF) == 0 || sec1 != sec2) _erase4kSector(sec2 << 12);

    // If we are about to write to a new sector, then erase it now.
    if ((mem & 0x0FFF) == 0) _erase4kSector(mem);

    SPI.beginTransaction(spiSet);
    _chipSelect();
    SPI.transfer(W25_CMD_WRITE_ENABLE);
    _chipDeSelect();
    _chipSelect();
    SPI.transfer(W25_CMD_PAGE_PROGRAM);
    _sendSeek(mem);

    uint8_t *addr = (uint8_t *)data;
    for (int i = 0; i < len; i++) SPI.transfer(*addr++);
    _chipDeSelect();
    SPI.endTransaction();
    _waitChipReady();

    #ifdef W25Q64FV_CFG_VERIFY
    boolean okay = true; int i;
    addr -= len;
    SPI.beginTransaction(spiSet);
    _chipSelect();
    SPI.transfer(W25_CMD_READ_ADDRESS);
    _sendSeek(mem);
    for (i = 0; i < len && okay; i++)
      if (*addr++ != SPI.transfer(0x00)) okay = false;
    _chipDeSelect();
    SPI.endTransaction();
    #ifdef W25Q64FV_CFG_DEBUG
    Serial.print("v:"); Serial.print(okay);
    Serial.print(" i="); Serial.println(i);
    #endif /*W25Q64FV_CFG_DEBUG*/
    if (!okay) writeBlock(data, mem, len); // Rewrite block.
    #endif /*W25Q64FV_CFG_VERIFY*/
  }
}
